"""
Wextracto uses `Function composition <http://en.wikipedia.org/wiki/Function_composition_%28computer_science%29>`_
as an easy way to build new functions from existing ones:

.. code-block:: pycon

    >>> from wex.composed import compose
    >>> def add1(x):
    ...     return x + 1
    ...
    >>> def mult2(x):
    ...     return x * 2
    ...
    >>> f = compose(add1, mult2)
    >>> f(2)
    6

Wextracto uses the pipe operator, ``|``, as a shorthand for function composition.

This shorthand can be a powerful technique for reducing boilerplate code when
used in combination with :class:`.Attributes`:

.. code-block:: python

    from wex.etree import css, text
    from wex.extractor import Attributes

    attrs = Attributes(
        title = css('h1') | text
        description = css('#description') | text
    )

"""

from itertools import chain
from functools import WRAPPER_ASSIGNMENTS, WRAPPER_UPDATES, partial as functools_partial
from six.moves import map as six_map


def compose(*functions):
    """ Create a :class:`.ComposedCallable` from zero more functions. """
    return ComposedCallable(*functions)


def composable(func):
    """ Decorates a callable to support function composition using ``|``.

    For example:

    .. code-block:: python

        @Composable.decorate
        def add1(x):
            return x + 1

        def mult2(x):
            return x * 2

        composed = add1 | mult2
    """
    return Composable.decorate(func)


class Composable(object):

    @classmethod
    def decorate(cls, func, **kw):
        name = getattr(func, '__name__', str(func))
        clsdict = dict(
            __call__=staticmethod(func),
            __doc__=getattr(func, '__doc__', None),
            __name__=getattr(func, '__name__', None),
            __module__=getattr(func, '__module__', None),
        )
        clsdict.update(kw)
        return type(name, (cls,), clsdict)()

    @classmethod
    def __getattr__(cls, name):
        return getattr(cls.__call__, name)

    @classmethod
    def __compose__(cls):
        return (cls.__call__,)

    def __or__(self, rhs):
        assert hasattr(rhs, '__call__')
        return compose(self, rhs)

    def __ror__(self, lhs):
        assert hasattr(lhs, '__call__')
        return compose(lhs, self)

    def __call__(self, arg):
        raise NotImplementedError


def flatten_composed_callables(functions):
    iterable = (getattr(f, 'functions', (f,)) for f in functions)
    return tuple(chain.from_iterable(iterable))


class ComposedCallable(Composable):
    """ A callable, taking one argument, composed from other callables.

    .. code-block:: python

        def mult2(x):
            return x * 2

        def add1(x):
            return x + 1

        composed = ComposedCallable(add1, mult2)

        for x in (1, 2, 3):
            assert composed(x) == mult2(add1(x))

    ComposedCallable objects are :func:`composable <wex.composed.composable>`.
    It can be composed of other ComposedCallable objects.  
    """

    def __init__(self, *functions):
        self.functions = flatten_composed_callables(functions)

    def __call__(self, arg, **kw):
        res = arg
        for func in self.functions:
            res = func(res, **kw)
        return res

    def __compose__(self):
        return self.functions

    def __repr__(self):
        return '<%s.%s%r>' % (self.__class__.__module__,
                              self.__class__.__name__,
                              self.functions)


def wraps(wrapped,
          assigned = WRAPPER_ASSIGNMENTS,
          updated = WRAPPER_UPDATES):
    defaults = {
        '__annotations__': {}
    }
    def decorator(wrapper):
        for attr in assigned:
            class_ = getattr(wrapped, '__class__', None)
            try:
                value = getattr(wrapped, attr)
            except AttributeError:
                try:
                    value = getattr(class_, attr)
                except AttributeError:
                    value = defaults[attr]
            setattr(wrapper, attr, value)
        for attr in updated:
            value = getattr(wrapped, attr, {})
            getattr(wrapper, attr).update(value)
        return wrapper
    return decorator


def partial(func, *args, **kwargs):
    return composable(functools_partial(func, *args, **kwargs))


def map(func):
    return partial(six_map, func)



""" Compatability fixes to make Python 2.7 look more like Python 3.

The general approach is to code using the common subset offered by 'six'.

The HTTPMessage class has a different interface.  This work-arounds makes the
Python 2.7 look enough like the Python 3 for the Wextracto code to work.
"""

import six

if six.PY2:

    from httplib import HTTPMessage

    def get_content_subtype(self):
        return self.getsubtype()
    HTTPMessage.get_content_subtype = get_content_subtype

    def get_content_charset(self):
        return self.getparam('charset')
    HTTPMessage.get_content_charset = get_content_charset


    def parse_headers(fp):
        return HTTPMessage(fp, 0)

else:

    from http.client import parse_headers  # pragma: no cover
    assert parse_headers                   # pragma: no cover



""" Helper functions for things that are iterable """

import wex.py2compat ; assert wex.py2compat
from itertools import chain, islice as islice_
from six import next, string_types
from six.moves import map, filter
from .composed import composable, wraps


class ZeroValuesError(ValueError):
    """ Zero values were found when at least one was expected. """


class MultipleValuesError(ValueError):
    """ More than one value was found when one or none were expected. """


# these are the types we do not want to iterate.
do_not_iter = tuple(string_types) + (dict, tuple)


def _do_not_iter_append(typeobj):
    # do_not_iter needs to be a tuple because we pass it to isinstance
    # but we want to append things so this makes it a little bit mutable
    global do_not_iter
    do_not_iter = do_not_iter + (typeobj,)


def should_iter(obj):
    return hasattr(obj, '__iter__') and not isinstance(obj, do_not_iter)


def should_iter_list(obj):
    return hasattr(obj, '__iter__') and not isinstance(obj, do_not_iter + (list,))


def walk(obj, should_iter=should_iter):

    def _walk(iterator):
        # pro-tip: step *into* this next line to debug generators
        for obj in iterator:
            if should_iter(obj):
                stack.append(iter(obj))
                return
            yield obj
        # iterator is now exhausted
        stack.pop()

    if not should_iter(obj):
        yield iter([obj])
        return

    # our stack of iterators - this is how we walk
    stack = [iter(obj)]

    # keep yielding generators until the stack is empty
    while stack:
        gen = _walk(stack[-1])
        yield gen
        # we expect that the generator will have been
        # exhausted by the time we get here, but we make
        # sure it is because otherwise the walk won't stop!
        for _ in gen:
            pass


@composable
def flatten(obj, should_iter=should_iter):
    """ Yield objects from all sub-iterables from obj. """
    return chain.from_iterable(walk(obj, should_iter))


@composable
def flatten_list(obj, should_iter=should_iter):
    """ Yield objects from all sub-iterables from obj. """
    return chain.from_iterable(walk(obj, should_iter_list))


def map_if_iter(func, should_iter=should_iter):
    @composable
    @wraps(func)
    def _map_if_iter(arg):
        if should_iter(arg):
            return map(func, arg)
        else:
            return func(arg)
    return _map_if_iter



def filter_if_iter(func):
    @composable
    def wrapper(arg):
        if not should_iter(arg):
            return arg
        else:
            return filter(func, arg)
    return wrapper


@composable
def first(iterable):
    """ Returns first item from an iterable.

    :param iterable: The iterable.

    If the iterable is empty then ``None`` is returned.
    """
    if not hasattr(iterable, '__iter__'):
        # turns out it isn't iterable after all
        return iterable
    i = iter(iterable)
    try:
        v0 = next(i)
    except StopIteration:
        raise ZeroValuesError
    return v0


@composable
def one(iterable):
    """ Returns an item from an iterable of exactly one element.

    If the iterable comprises zero elements then :exc:`.ZeroValuesError` is
    raised.
    If the iterable has more than one element then :exc:`.MultipleValuesError` is
    raised.
    """
    if not hasattr(iterable, '__iter__'):
        # turns out it isn't iterable after all
        return iterable
    i = iter(iterable)
    v0 = first(i)
    try:
        next(i)
        raise MultipleValuesError()
    except StopIteration:
        pass
    return v0


@composable
def one_or_none(iterable):
    """ Returns one item or ``None`` from an iterable of length one or zero.

    If the iterable is empty then ``None`` is returned.

    If the iterable has more than one element then :exc:`.MultipleValuesError` is
    raised.
    """
    try:
        return one(iterable)
    except ZeroValuesError:
        return None


def islice(*islice_args):
    """ Returns a function that will perform ``itertools.islice`` on its input. """
    @composable
    def islice(iterable):
        return islice_(iterable, *islice_args)
    return islice



import sys
import re
from .composed import composable
from .iterable import flatten

for name in dir(re):
    if name.isupper() and not name.startswith('_'):
        setattr(sys.modules[__name__], name, getattr(re, name))


def re_group(pattern, group=1, flags=0):
    """
    Returns a :mod:`composable <wex.composed>` callable that 
    extract the specified group using a regular expression.

    :param pattern: The regular expression.
    :param group: The group from the `MatchObject <https://docs.python.org/2/library/re.html#re.MatchObject.group>`_.
    :param flags: Flags to use when compiling the 
                        `pattern <https://docs.python.org/2/library/re.html#re.compile>`_.
    """
    compiled = re.compile(pattern, flags)
    @composable
    def regroup(src):
        for string in flatten(src):
            for match in compiled.finditer(string):
                yield match.group(group)
    return regroup


def re_groupdict(pattern, flags=0):
    """
    Returns a :mod:`composable <wex.composed>` callable that 
    extract the a group dictionary using a regular expression.

    :param pattern: The regular expression.
    :param flags: Flags to use when compiling the 
                        `pattern <https://docs.python.org/2/library/re.html#re.compile>`_.
    """
    compiled = re.compile(pattern, flags)
    compiled = re.compile(pattern, flags)
    @composable
    def redict(src):
        for string in flatten(src):
            for match in compiled.finditer(string):
                yield match.groupdict()
    return redict



""" Functions for getting responses for HTTP urls. """

from __future__ import unicode_literals, print_function
import wex.py2compat ; assert wex.py2compat
import io
import requests
from gzip import GzipFile
from .readable import ChainedReadable


GZIP_MAGIC = b'\x1f\x8b'
CRLF = '\r\n'
timeout = 30.0


def request(url, method, session=None, **kw):
    """ Makes an HTTP request following redirects. """

    if session is None:
        session = requests.Session()
        session.stream = True

    decode_content = kw.get('decode_content', True)

    response = session.request(
        method.name,
        url,
        params=method.args.get('params', None),
        data=method.args.get('data', None),
        headers=method.args.get('headers', None),
        cookies=method.args.get('cookies', None),
        timeout=timeout,
        allow_redirects=False,
    )
    yield readable_from_response(response, url, decode_content)

    redirects = session.resolve_redirects(response, response.request,
                                          stream=True, timeout=timeout)
    for redirect in redirects:
        yield readable_from_response(redirect, url, decode_content)


def readable_from_response(response, url, decode_content=True):
    """ Make an object that is readable by `Response`.from_file. """

    headers = io.TextIOWrapper(io.BytesIO(), encoding='utf-8', newline='\n')
    protocol = 'HTTP'
    version = '{:.1f}'.format(response.raw.version / 10.0)
    code = response.status_code
    reason = response.reason
    status_line = format_status_line(protocol, version, code, reason)
    response.raw.decode_content = decode_content
    magic_bytes = response.raw.read(2)
    headers.write(status_line)
    for name, value in response.headers.items():
        headers.write(format_header(name.capitalize(), value))
    headers.write(format_header('X-wex-request-url', url))
    if response.url != url:
        # the URL for the response is not the same as the requested URL
        headers.write(format_header('X-wex-url', response.url))
    if magic_bytes == GZIP_MAGIC:
        headers.write(format_header('X-wex-has-gzip-magic', '1'))
    headers.write(CRLF)
    headers.seek(0)

    return ChainedReadable(headers.detach(), io.BytesIO(magic_bytes), response.raw)


def decode(src):
    content_encoding = src.headers.get('content-encoding', '')
    has_gzip_magic = (src.headers.get('X-wex-has-gzip-magic', '0') == '1')
    gzip = (
        src.headers.get_content_subtype() == 'x-gzip' or
        content_encoding == 'x-gzip' or
        (content_encoding == 'gzip' and has_gzip_magic)
    )
    if gzip:
        src.seek(0)
        return GzipFile(fileobj=src.fp, mode='rb')
    return src


format_status_line = '{}/{} {} {}\r\n'.format
format_header = '{}: {}\r\n'.format



""" Extracted data values are represented with tab-separated fields.
The right-most field on each line is the value, all preceding fields
are labels that describe the value.
The labels and the value are all JSON encoded.

So for example, a value 9.99 with a labels ``product`` and ``price`` would 
look like::

    "product"\t"price"\t9.99\n

And we could decode this line with the following Python snippet:

.. code-block:: pycon

    >>> import json
    >>> line = '"product"\\t"price"\\t9.99\\n'
    >>> [json.loads(s) for s in line.split('\\t')]
    [u'product', u'price', 9.99]

Using tab-delimiters is convenient for downstream processing using Unix 
command line tools such as :command:`cut` and :command:`grep`.
"""

import sys
from json import JSONEncoder
from itertools import product
from functools import partial
from operator import itemgetter
from six import PY2, text_type, reraise
from six.moves import map
import logging; logger = logging.getLogger(__name__)
from .iterable import walk, flatten, do_not_iter


TAB = '\t'
NL = '\n'

if PY2:
    JSONEncoder = partial(JSONEncoder, encoding='UTF-8')

json_encode = JSONEncoder(
    skipkeys=False,
    check_circular=True,
    allow_nan=True,
    indent=None,
    separators=(',', ':'),
    default=None,
    sort_keys=True,
    # may need to make this an argument at some point,
    # but for now let's assume UTF-8 is ok on the output.
    ensure_ascii=False,
).encode


def encode(obj):
    try:
        return json_encode(obj)
    except TypeError:
        return '#' + text_type(repr(obj)) + '!'


def should_iter_unless_list(obj):
    # a list is a reasonable value type so don't flatten it
    return hasattr(obj, '__iter__') and not isinstance(obj, do_not_iter + (list,))


class Value(tuple):

    exit_on_exc = False
    debug_on_exc = False

    value = property(itemgetter(-1))
    labels = property(itemgetter(slice(0, -1)))

    def __new__(cls, value=(None,)):
        if not isinstance(value, tuple):
            value = (value,)
        return super(Value, cls).__new__(cls, value)

    def text(self):
        """ Returns the text this value as a labelled JSON line. """
        iterables = [map(encode, flatten(label)) for label in self.labels]
        iterables.append(map(encode, flatten(self.value, should_iter_unless_list)))
        for fields in product(*iterables):
            yield TAB.join(fields) + NL

    def label(self, *labels):
        """ Adds zero or more labels to this value. """
        return self.__class__(tuple(labels) + self)



def yield_values(extract, *args, **kw):
    """ Yields ``Value`` objects extracted using ``extract``. """
    exc_info = ()

    try:
        returned = extract(*args, **kw)
        for walker in walk(returned, should_iter_unless_list):
            for value in walker:
                yield Value(value)
    except Exception as exc:
        exc_info = sys.exc_info()
        yield Value(exc)

    if any(exc_info) and (Value.exit_on_exc or Value.debug_on_exc):
        if Value.debug_on_exc:
            import pdb
            pdb.post_mortem(exc_info[2])
        else:
            reraise(exc_info[0], exc_info[1], exc_info[2])



import os
import errno
import json
import logging
from operator import attrgetter, methodcaller
from hashlib import md5
from contextlib import contextmanager
from six import text_type, string_types, next
from six.moves import filter
from six.moves.urllib_parse import (urlparse,
                                    urlunparse,
                                    parse_qs,
                                    parse_qsl,
                                    urlencode,
                                    unquote,
                                    quote)
from pkg_resources import iter_entry_points
from publicsuffix import PublicSuffixList

from .composed import composable
from .iterable import map_if_iter
from .value import json_encode

logger = logging.getLogger(__name__)

DEFAULT_METHOD = 'get'

if hasattr(os, 'pathconf'):
    PC_NAME_MAX = os.pathconf(os.path.dirname(__file__), 'PC_NAME_MAX')
else:
    PC_NAME_MAX = 255  # pragma: no cover


@contextmanager
def eexist_is_ok():
    try:
        yield
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise


class Method(object):
    """ Method objects 'get' responses from a url.

        The Method object looks-up the correct implementation based on its
        name and the scheme of the url.

        The default method name is 'get'. Other method names can be specified
        in the fragment of the url.
    """
    def __init__(self, scheme, name, args=None):
        self.scheme = scheme
        self.name = name
        self.args = args or {}

    @property
    def group(self):
        return 'wex.method.{}'.format(self.scheme)

    def get(self, url, **kw):
        """ Get responses for 'url'. """
        entry_points = iter_entry_points(self.group, self.name)
        try:
            ep = next(entry_points)
        except StopIteration:
            raise ValueError("Missing method '%s' in '%s'" % (self.name, self.group))
        method = ep.load()
        return method(url, self, **kw)


class URL(text_type):
    """ URL objects. """

    def __new__(cls, urlstring):
        url = super(URL, cls).__new__(cls, urlstring)
        url.parsed = urlparse(url)
        return url


    @property
    def fragment_dict(self):
        """ Client side data dict represented as JSON in the fragment. """
        if not self.parsed.fragment:
            return {}

        if self.parsed.fragment.startswith('%7B'):
            fragment = unquote(self.parsed.fragment)
        elif not self.parsed.fragment.startswith('{'):
            return {}
        else:
            fragment = self.parsed.fragment

        try:
            data = json.loads(fragment)
            if not isinstance(data, dict):
                data = {}
        except ValueError as exc:
            logger.error("%s. Unable to parse %r", exc, fragment)
            data = {}

        return data

    def update_fragment_dict(self, **kw):
        fragment_dict = dict(self.fragment_dict)
        fragment_dict.update(kw)
        fragment = json_encode(fragment_dict)
        return self.__class__(urlunparse(self.parsed._replace(fragment=fragment)))

    @property
    def method(self):
        """ The `Method` for this URL. """
        if not self.parsed.scheme:
            raise ValueError("URL has no scheme")

        method = self.fragment_dict.get('method', DEFAULT_METHOD)

        if isinstance(method, string_types):
            return Method(self.parsed.scheme, method, {})

        try:
            ((name, args),) = method.items()
        except:
            raise ValueError("invalid method %r" % method)
        return Method(self.parsed.scheme, name, args)

    def get(self, **kw):
        """ Get `url` using the appropriate `Method`. """
        return self.method.get(self, **kw)

    def mkdirs(self, top):
        dirpath = top
        for dirname in self.dirnames():
            dirpath = os.path.join(dirpath, dirname)
            with eexist_is_ok():
                os.mkdir(dirpath)
        return dirpath

    def dirnames(self):
        encoded = self.encode('utf-8')
        hexdigest = md5(encoded).hexdigest()
        names = [self.parsed.scheme, self.parsed.netloc]
        names.extend(self.parsed.path.split('/'))
        if self.parsed.query:
            names.extend(self.parsed.query.split('&'))
        names.append(hexdigest)
        return [quote(name, safe='')[:PC_NAME_MAX] for name in names]



#
# URL related composable helpers
#============================================================


public_suffix_list = PublicSuffixList()


@composable
@map_if_iter
def url(obj):
    return getattr(obj, 'url', obj)

parse_url = url | map_if_iter(urlparse)
url_query = parse_url | map_if_iter(attrgetter('query'))
url_path = parse_url | map_if_iter(attrgetter('path'))
url_hostname = parse_url | map_if_iter(attrgetter('hostname'))
url_query_dict = url_query | map_if_iter(parse_qs)
url_query_list = url_query | map_if_iter(parse_qsl)

def url_query_param(name, default=[]):
    return url_query_dict | map_if_iter(methodcaller('get', name, default))


def filter_url_query(*names, **kw):

    names = set(names)
    exclude = kw.pop('exclude', False)

    def included(pair):
        return pair[0] in names

    def excluded(pair):
        return pair[0] not in names

    if exclude:
        pred = excluded
    else:
        pred = included

    @composable
    @map_if_iter
    def url_query_filter(obj):
        parsed = parse_url(obj)
        qsl = list(filter(pred, parse_qsl(parsed.query)))
        filtered_query = urlencode(qsl)
        return urlunparse(parsed._replace(query=filtered_query))

    return url_query_filter


strip_url_query = filter_url_query()


@composable
@map_if_iter
def public_suffix(src):
    return public_suffix_list.get_public_suffix(url_hostname(src) or src)



from __future__ import unicode_literals
from wex.composed import map
from wex.iterable import flatten
from operator import methodcaller


space_join = ' '.join


def split(*split_args):
    split_func = methodcaller('split', *split_args)
    return map(split_func)


# with default arguments splits whitespace
split_ws = split()


def norm_ws_1(s):
    return space_join(s.split())


norm_ws = flatten | map(norm_ws_1)


def partition(separator, **kw):
    norm = kw.pop('norm', norm_ws_1)
    norm_head = kw.pop('norm_head', norm)
    norm_tail = kw.pop('norm_tail', norm)
    def partition_1(s):
        head, sep, tail = s.partition(separator)
        if norm_head:
            head = norm_head(head)
        if norm_tail:
            tail = norm_tail(tail)
        if sep:
            yield (head, tail)
    return flatten | map(partition_1)



from __future__ import unicode_literals, print_function, absolute_import
from tempfile import TemporaryFile
from shutil import copyfileobj
from io import BytesIO
from six.moves.urllib.response import addinfourl
from six.moves.http_client import BadStatusLine
from .py2compat import parse_headers
from .cache import Cache
from .value import yield_values
from .iterable import _do_not_iter_append


DEFAULT_READ_SIZE = 2**16  # 64K
MAX_IN_MEMORY_SIZE = 2**29      # 512M



class Response(addinfourl):
    """ A urllib2 style Response with some extras. 

        :param content: A file-like object containing the response content.
        :param headers: An HTTPMessage containing the response headers.
        :param url: The URL for which this is the response.
        :param code: The status code recieved with this response.
        :param protocol: The protocol received with this response.
        :param version: The protocol version received with this response.
        :param reason: The reason received with this response.
        :param request_url: The URL requested that led to this response.
    """

    def __init__(self, content, headers, url, code=None, **kw):
        addinfourl.__init__(self, content, headers, url, code)
        self.request_url = kw.pop('request_url', None)
        self.protocol = kw.pop('protocol', None)
        self.version = kw.pop('version', None)
        self.reason = kw.pop('reason', None)
        if kw:
            raise ValueError("unexpected keyword arguments %r" % kw.keys())

    def seek(self, offset=0, whence=0):
        """ Seek the content file position.

            :param int offset: The offset from whence.
            :param int whence: 0=from start,1=from current position,2=from end
        """
        self.fp.seek(offset, whence)
        return self.fp

    @classmethod
    def values_from_readable(cls, extractor, readable):
        response = cls.from_readable(readable)
        with Cache():
            for value in yield_values(extractor, response):
                yield value

    @classmethod
    def from_readable(cls, readable):
        status_line = readable.readline()
        protocol, version, code, reason = cls.parse_status_line(status_line)
        headers = parse_headers(readable)
        request_url = headers.get('X-wex-request-url')
        url = headers.get('X-wex-url', request_url)
        content = cls.content_file(readable, headers)
        return Response(content, headers, url,
                                          code=code,
                                          protocol=protocol.decode('UTF-8'),
                                          version=version,
                                          reason=reason.decode('UTF-8'),
                                          request_url=request_url)

    @staticmethod
    def parse_status_line(status_line, field_defaults=['']*3):
        fields = status_line.rstrip(b'\r\n').split(None, 2) + field_defaults
        protocol_version, code, reason = fields[:3]

        # status code is always an integer
        if not code.isdigit():
            raise BadStatusLine(status_line)
        code = int(code)

        protocol, _, version = protocol_version.partition(b'/')

        # version is a tuple of integers
        try:
            version = tuple(map(int, version.split(b'.')))
        except ValueError:
            raise BadStatusLine(status_line)

        return protocol, version, code, reason

    @classmethod
    def content_file(cls, response_file, headers):
        try:
            content_length = int(headers.get('content-length', 0))
        except ValueError:
            content_length = 0

        size_with_content_length = min(content_length + 1, MAX_IN_MEMORY_SIZE)
        read_size = max(size_with_content_length, DEFAULT_READ_SIZE)
        buf = response_file.read(read_size)
        if len(buf) < read_size:
            # We've managed to read all the content in one go
            content_file = BytesIO(buf)
        else:
            content_file = TemporaryFile()
            content_file.write(buf)
            copyfileobj(response_file, content_file)
            content_file.seek(0)

        return content_file


# Response supports __iter__ because that is normal for file-like objects
# but by default we don't want respones to be iterated when flattening or
# when composable helpers are trying to work out whether to map or not.
_do_not_iter_append(Response)

